
#include <UIPEthernet.h>
#include <SoftwareSerial.h>


/**************************** CONFIG ****************************/
#define DEBUG 0
#define STATIC 0

// define switches IO Ports
#define SERVER1 3
#define SERVER2  4
#define SERVER3  5

// speed and buffer settings
#define TELNET_SPEED 9600
#define SERIAL_BUFFER_SIZE 64
#define HTTP_BUFFER_SIZE 128   // also max number of iteration to parse query parts

// custom debug macro
#if DEBUG == 1
  #define _MSGLN(str) Serial.println(F(str))
  #define _MSG(str)   Serial.print(F(str))
  #define _VARLN(str) Serial.println(str)
  #define _VAR(str)   Serial.print(str)
#else
  #define _MSGLN(str) 
  #define _MSG(str) 
  #define _VARLN(str)
  #define _VAR(str)
#endif

// ethernet mac address
static uint8_t mac[6] = {0xfe,0x68,0x70,0x2a,0x22,0x30};

#if STATIC == 1
// ethernet interface ip address
static IPAddress ip(192,168,1,100);  
// gateway ip address
static IPAddress gwip(192,168,1,254);
#endif

// Network Listeners
EthernetServer HttpServer   = EthernetServer(80);
EthernetServer TelnetServer = EthernetServer(23);
char HttpReqBuff[HTTP_BUFFER_SIZE];

// Serial Listeners
#define SERVER1_RX 15
#define SERVER1_TX 14
#define SERVER2_RX  17
#define SERVER2_TX  16
#define SERVER3_RX  19
#define SERVER3_TX  18

SoftwareSerial *CurrentSoftSerial = NULL;
int CurrentRXPin;
int SerialBufferPos = 0;
char SerialBuffer[SERIAL_BUFFER_SIZE] = "\0";
int NbWrites = 0;
/***************************************************************/


/************************ Main function ************************/
void setup()
{
#if DEBUG == 1
  Serial.begin(9600);
#endif
    
  _MSGLN("starting system...");
  initSwitches();
  initNetwork(); 
  initServers();

  setCurrentSoftSerial(SERVER1_RX, SERVER1_TX);
  _MSGLN("Waiting for connections");
}


void loop()
{  
  handleTelnet();
  handleHttp();
}
/**************************************************************/


/********************** hardware related ************************/
// Set switches PIN
void initSwitches() {
  
  _MSGLN("initialize switches...");
  
  // Ensure off by default
  digitalWrite(SERVER3,  HIGH);
  digitalWrite(SERVER2,  HIGH);
  digitalWrite(SERVER1, HIGH);

  pinMode(SERVER3, OUTPUT);
  pinMode(SERVER2, OUTPUT);
  pinMode(SERVER1,OUTPUT);
}
/**************************************************************/



/********************** Serial related ************************/
void setCurrentSoftSerial(int RX, int TX) {
  CurrentRXPin = RX;
  if(CurrentSoftSerial) delete(CurrentSoftSerial);

  CurrentSoftSerial = new SoftwareSerial(RX, TX);
  CurrentSoftSerial->begin(TELNET_SPEED);

  _MSG("Software serial changed : RX =  ");
  _VARLN(CurrentRXPin);
}
/***************************************************************/



/********************** Network related ************************/
void initNetwork() {
  int conn = 1;
  
  _MSGLN("Initialize network...");

#if STATIC == 1
  Ethernet.begin(mac, ip, gwip, gwip);  // initialize Ethernet device with static IP
//  Ethernet.begin(mac, ip);
#else
  conn = Ethernet.begin(mac);           // initialize Ethernet with DHCP
#endif

  if (conn == 0) {
    _MSGLN("  Failed to confgure Ethernet using DHCP");
    return;
  }

  _MSG("  IP Address : ");
  _VARLN(Ethernet.localIP());
}

// start services listeners
void initServers() {
  _MSGLN("Initialize Servers...");
  HttpServer.begin();
  TelnetServer.begin();
}
/***************************************************************/



/************************ HTTP related *************************/
void handleHttp() {
  EthernetClient client = HttpServer.available();
  
  if (client) {
    _MSGLN("HTTP connection detected !");
    char c;
    int i;

    // first loop : ignore request type
    for(i=0, c = client.read(); client.available() && c != ' ' && i < HTTP_BUFFER_SIZE ; i++, c = client.read()) {
      _MSG("1 read : ") ; _VARLN(c);
    }
    
    // second loop : get the URI
    for(i=0, c = client.read(); client.available() && c != ' ' && c != '\n' && i < HTTP_BUFFER_SIZE ; i++, c = client.read()) {
      _MSG("2 read ") ; _VAR(i) ; _MSG(" : ") ; _VARLN(c);
      HttpReqBuff[i] = c;
    }
    
    HttpReqBuff[i] = '\0';
    _MSG("Final string : "); _VARLN(HttpReqBuff);
    
    handleHttpAction();  
    handleHttpHome(&client);
    delay(1);
    client.stop();
    _MSGLN("HTTP connection closed");
  }
}


// turn switches ON or OFF or set current serial connection depending on HTTP query
static void handleHttpAction() {

  _MSGLN("handle http action");
   
  if(!strcmp_P(HttpReqBuff, PSTR("/SERVER1/ON"))  || !strcmp_P(HttpReqBuff, PSTR("/ALL/ON")))  digitalWrite(SERVER1, LOW);
  if(!strcmp_P(HttpReqBuff, PSTR("/SERVER1/OFF")) || !strcmp_P(HttpReqBuff, PSTR("/ALL/OFF"))) digitalWrite(SERVER1, HIGH);
  
  if(!strcmp_P(HttpReqBuff, PSTR("/SERVER2/ON"))  || !strcmp_P(HttpReqBuff, PSTR("/ALL/ON")))  digitalWrite(SERVER2, LOW);
  if(!strcmp_P(HttpReqBuff, PSTR("/SERVER2/OFF")) || !strcmp_P(HttpReqBuff, PSTR("/ALL/OFF"))) digitalWrite(SERVER2, HIGH);
  
  if(!strcmp_P(HttpReqBuff, PSTR("/SERVER3/ON"))  || !strcmp_P(HttpReqBuff, PSTR("/ALL/ON")))  digitalWrite(SERVER3, LOW);
  if(!strcmp_P(HttpReqBuff, PSTR("/SERVER3/OFF")) || !strcmp_P(HttpReqBuff, PSTR("/ALL/OFF"))) digitalWrite(SERVER3, HIGH);
  
  if(!strcmp_P(HttpReqBuff, PSTR("/SERVER1/TELNET"))) { setCurrentSoftSerial(SERVER1_RX, SERVER1_TX);return; }
  if(!strcmp_P(HttpReqBuff, PSTR("/SERVER2/TELNET"))) { setCurrentSoftSerial(SERVER2_RX,  SERVER2_TX); return; }
  if(!strcmp_P(HttpReqBuff, PSTR("/SERVER3/TELNET"))) { setCurrentSoftSerial(SERVER3_RX,  SERVER3_TX); return; }
}


// Send http response with switches statuses
void handleHttpHome(EthernetClient *client) {

  _MSGLN("sending home page");

  // send HTTP headers
  client->println(F("HTTP/1.0 200 OK"));
  client->println(F("Content-Type: text/xml"));
  client->println(F("Cache-Control: no-cache, no-store, must-revalidate"));
  client->println(F("Pragma: no-cache"));
  client->println(F("Expires: 0"));
  client->println(F("Connection: close")); 
  client->println();
  
  // send content
  client->println(F("<ROOT>"));
  client->print(F("<SERVER1>")); client->print(!digitalRead(SERVER1));client->println(F("</SERVER1>"));
  client->print(F("<SERVER2>")); client->print(!digitalRead(SERVER2)) ;client->println(F("</SERVER2>"));
  client->print(F("<SERVER3>")); client->print(!digitalRead(SERVER3)) ;client->println(F("</SERVER3>"));
  client->print(F("<SERIAL>"));
  if      (CurrentRXPin == SERVER1_RX) client->print(F("SERVER1"));
  else if (CurrentRXPin == SERVER2_RX) client->print(F("SERVER2"));
  else if (CurrentRXPin == SERVER3_RX) client->print(F("SERVER3"));
  client->println(F("</SERIAL>"));
  client->println(F("</ROOT>"));
  
}
/***************************************************************/


/*********************** Telnet related ************************/
void handleTelnet() { 
  
  EthernetClient client = TelnetServer.available();
  char c;
  
  // if something was typed through telnet, send it to serial connection
  if(client) { _MSGLN("send serial char"); CurrentSoftSerial->write(client.read()); }
 
  // handle incomming byte from serial connection and write it to telnet connection
  // buffering is mandatory to not loose any byte.
  while(CurrentSoftSerial->available()) {
    
    c = CurrentSoftSerial->read();
    if(c < 0) continue;
    SerialBuffer[SerialBufferPos] = c;
     
    //_MSG("Getting char "); _VAR(SerialBufferPos); _MSG(" :") ; _VARLN(SerialBuffer[SerialBufferPos-1]);
    // send data as soon as buffer is full
    if(SerialBufferPos==SERIAL_BUFFER_SIZE-1)
        WriteSerialBufferToTelnet(SERIAL_BUFFER_SIZE);
    
    SerialBufferPos++;
  }
  
  // handle case where no more data available, but still some chars left in buffer
  WriteSerialBufferToTelnet(SerialBufferPos);
}


void WriteSerialBufferToTelnet(int sizeToWrite) {
  if(sizeToWrite > 0 && sizeToWrite <= SERIAL_BUFFER_SIZE) {
    TelnetServer.write(SerialBuffer, sizeToWrite);
    _VARLN(SerialBuffer);
    SerialBufferPos=0;
  }
}
/***************************************************************/
